program banktools;

{
    Tools for manipulation bank of questions for tests.
    Copyright (C) 2009-2019  Artyomov Alexander and authors of content listed below

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
}

(* (a) Sergey V Kopylov, Alexander N Artyomov (c), (cl) 2009.
   License: GNU GPL. All rights reserved.
   Version: 0.1 (initial) *)

(* (a) Sergey V Kopylov, Alexander N Artyomov, Stanislav A Kozlovsky (c), (cl) 2010.
   License: GNU GPLv3. All rights reserved.
   Version: 0.2 *)

(* (a) Sergey V Kopylov, Alexander N Artyomov (c), (cl) 2010.
   License: GNU GPLv3. All rights reserved.
   Version: 0.3 *)

(* (a) Sergey V Kopylov, Alexander N Artyomov, Stanislav A Kozlovsky (c), (cl) 2010.
   License: GNU GPLv3. All rights reserved.
   Version: 0.4 *)

{$MODE OBJFPC}    // Режим FreePascal    с объектами
{$LONGSTRINGS ON} // Включаем поддержку длинных строк
{$RANGECHECKS ON} // Включаем проверку диапазонов
{$SMARTLINK ON}   // Включаем "умную" линковку
{$INLINE ON}      // Включаем inline функции

uses SysUtils, Classes, Unix;

type

     subjs = 1..15; // тем
     anss  = 1..4;  // вариантов ответа
     trs   = 1..3;  // сложность
     aanss = array[anss] of widestring; // Массив для вариантов ответа
     aoL   = array of LongInt; // Для разных применений

     TQuest = class(TObject)  // Вопрос
        tr  : trs;    // Сложность
        subj: subjs;  // Тема
        q   : widestring; // Текст вопроса
        a   : aanss;  // Варианты ответа
        tru : anss;   // Верный ответ
        sol : widestring; // Решение
     public
     function OutTeXforPB(num : LongInt) : widestring; // Вывод вопросов банка для его печати
     function Equal(qu : TQuest) : boolean; // Вопросы одинаковы
     end;
     TArrayOfTQuest = array of TQuest;

     variants  = 0..2; // число вариантов (билетов) в пакете

const
     subjsnames : array[subjs] of widestring =
     ('М1','М2','М3','М4','Т1','Т2','Т3','Т4','Э1','Э2','Э3','Э4','О','А','Я');

     lots : array[subjs] of LongInt = (1,1,1,1,1,1,1,1,1,1,1,1,1,1,1); // количество задач в пакете на соответствующую тему
     variants_count = High(variants); // число вариантов (билетов) в пакете
     packetcnt = 120; // число пакетов


type
     TBank = class(TObject)
     fqa : TArrayOfTQuest; // Динамический массив вопросов
     public
     property QA : TArrayOfTQuest read fqa write fqa;
     function LoadFromFile(fn : widestring) : boolean; // Загрузка массива вопросов из файла банка вопросов
     procedure PrintBank(fn : utf8string);
     function SaveToFile(fn : UTF8string) : boolean;
     function SaveToFile1Part(fn : UTF8string) : boolean;
     function ExcludeToFile(fn : UTF8string; var b : TBank) : boolean;
     end;

procedure RCleanStr(var s : widestring); var f : LongInt;
begin for f := Length(s) downto 1 do if s[f] > #32 then begin SetLength(s, f); Exit; end; s := ''; end;
procedure LCleanStr(var s : widestring); var f, l : LongInt;
begin l := Length(s); for f := 1 to l do if s[f] > #32 then begin s := Copy(s, f, l - f + 1); Exit; end;  s := ''; end;
function SpaceOne(s : UTF8string) : utf8string; // замена ряда пробелов на один
  var
      tmp : UTF8string = '';
      f   : LongInt;
      ss  : boolean = false;
begin
for f := 1 to Length(s) do begin
    if s[f] = ' ' then begin if not ss then tmp := tmp + ' '; ss := true; continue; end else ss := false;
    if not ss then tmp := tmp + s[f];
end;
Exit(tmp);
end;
procedure NeatStr(var s : widestring); // очистка строки
begin RCleanStr(s); LCleanStr(s); s := SpaceOne(s); end;

function TBank.LoadFromFile(fn : widestring) : boolean; // Загрузка массива вопросов из файла банка вопросов
  var
          fp : Text;
           s : utf8string;
       state : LongInt = 0; // Состояние обработки
           q : TQuest;      // Временная переменная
           p : LongInt = 0; // Хранит позицию подстроки в строке
        lcnt : LongInt = 0; // Счётчик строк "банка"
        acnt : LongInt = 0; // Счётчик ответов
        stmp : widestring;      // Временная строка
               procedure addtoarray;
                 var
                     f : LongInt;
               begin
                    // очищаем строки от пустот по краям
                    NeatStr(q.q); NeatStr(q.sol); for f := 1 to High(q.a) do NeatStr(q.a[f]);
                    // увеличиваем размер массива и присваиваем временную переменную
                    SetLength(fqa, Length(qa) + 1); qa[High(qa)] := TQuest.Create; qa[High(qa)] := q;
                    // сбрасываем строковые значения на пустые, создавая новый объект
                    q := TQuest.Create; // создаём новенький класс с тем же указателем - старый ушёл в массив
               end;
begin

q := TQuest.Create;

Assign(fp, fn);
WriteLn(stderr, 'Подготавливаем файл банка заданий к чтению...'); WriteLn(stderr);
ReSet(fp);
while not eof(fp) do begin
ReadLn(fp, s); inc(lcnt); // Читаем строку и считаем её

case state of
0: begin // Пока нет никаких секций (до вопросов) - ждём их
   p := Pos('\section{', s);
   if p > 0 then begin
      WriteLn(stderr, 'На ', lcnt, ' строке найдена \section{ ...');
      q.subj := StrToIntDef(Copy(s, p + 9, Pos('}', s) - (p + 9)), 0); // Если ошибка набора - сработает rangechecks
      WriteLn(stderr, '  Ok, Тема = ', q.subj); WriteLn(stderr);
      state := 1;
   end; {end if}
end;
1: begin // Аккумулируем текст задания и ждём субсекцию
   p := Pos('\subsection{', s);
   if p > 0 then begin
      WriteLn(stderr, 'На ', lcnt, ' строке найдена \subsection{ ...');
      inc(acnt);

      stmp := Copy(s, p + 12, Pos('}', s) - (p + 12));

      case acnt of
      1..High(anss): begin
                     if stmp  = '+' then begin q.tru := acnt; WriteLn(stderr, '  Вариант ', q.tru, ' установлен как верный'); end;
      end;
      else
        WriteLn(stderr, '  Устанавливаем сложность ...');
        q.tr := StrToIntDef(stmp, 0); // Если ошибка набора - сработает rangechecks
        WriteLn(stderr, '  Ok. Сложность ', q.tr);
        state := 2;
        acnt := 0;
      end; {end select}
   end else begin
                   // попутно накапливаем текст вопрсов
                   if acnt = 0 then q.q := q.q + #10 + s else q.a[acnt] := q.a[acnt] + #10 + s; 
   end; {end if}
end;
2: begin // Ждём новую секцию, а пока накапливаем решение
   p := Pos('\section{', s);
   if p > 0 then begin
        AddToArray;
      WriteLn(stderr, 'На ', lcnt, ' строке найдена \section{ ...');
      q.subj := StrToIntDef(Copy(s, p + 9, Pos('}', s) - (p + 9)), 0); // Если ошибка набора - сработает rangechecks
      WriteLn(stderr, '  Ok, Тема = ', q.subj); WriteLn(stderr);
      state := 1;
   end else begin
      q.sol := q.sol + #10 + s; // накапливаем решение
   end; {end if}
end;


end; {end select}

end; {wend}
Close(fp);

// если "банк" закончился, а состояние чтения решения - иначе ошибка структуры "банка"
if state = 2 then AddToArray else WriteLn(stderr, 'Ошибка структуры в последнем вопросе.'); 
Exit(false); // Ok. Ob.
end;

function TQuest.OutTeXforPB(num : LongInt) : widestring;
  var
      f   : LongInt;
      tmp : widestring = '';
begin

for f := 1 to High(a) do if f <> tru then
    tmp := tmp + '\hspace{4pt}' + Char(LongInt(Char('A')) + f - 1) + ')\hspace{1pt}' + a[f] + ' '
    else tmp := tmp + '\hspace{4pt}\textbf{' + Char(LongInt(Char('A')) + f - 1) + ')}{$\bigstar$}\hspace{1pt}' + a[f] + ' ';

Exit('\textbf{' + IntToStr(num) + '.} ' + #10 + 'Раздел: ' + subjsnames[subj] + ', Сложность: ' + IntToStr(tr) + #10#10 + q + #10#10 +
'\textbf{Варианты ответов:} ' + #10 +
tmp + #10#10 + '\smallskip \hrule \smallskip'  + #10#10#10);
end;

function TQuest.Equal(qu : TQuest) : boolean;
  var
      f : LongInt;
begin
for f := Low(aanss) to High(aanss) do if qu.a[f] <> a[f] then Exit(false);
Exit(
//(tru = qu.tru) and
(q = qu.q)
);
end;

procedure TBank.PrintBank(fn : utf8string);
  var     fp : Text; f : LongInt;
begin
Assign(fp, fn);
ReWrite(fp);

WriteLn(fp, '\documentclass[11pt]{article}');
WriteLn(fp, '\usepackage{mathtext}');
WriteLn(fp, '\usepackage[utf8x]{inputenc}');
WriteLn(fp, '\usepackage{graphicx}');
WriteLn(fp, '\usepackage[T2A]{fontenc}');
WriteLn(fp, '\usepackage[english,russian]{babel}');
WriteLn(fp, '\usepackage{amsmath,amstext,amssymb,array,textcomp}');
WriteLn(fp, '\usepackage[papersize={210mm,297mm},left=1cm,right=1cm,top=1cm,bottom=1cm,bindingoffset=0cm]{geometry}');
WriteLn(fp, '\pagestyle{empty} % Убираем нумерацию страниц');
WriteLn(fp, '\parindent=0pt % Выключам абзацный отступ');
WriteLn(fp, '');

WriteLn(fp, '\begin{document}');
WriteLn(fp, '');

for f := 0 to High(qa) do WriteLn(fp, qa[f].OutTeXforPB(f+1));

WriteLn(fp, '');
WriteLn(fp, '\end{document}');
Close(fp);
end;

function TBank.ExcludeToFile(fn : UTF8string; var b : TBank) : boolean;
  var
      f, ff, fff : LongInt;
      eq         : boolean;
      fp         : Text;
begin
Assign(fp, fn);
ReWrite(fp);

  for f := Low(qa) to High(qa) do begin
  eq := false;
  
  
  for ff := Low(b.qa) to High(b.qa) do begin
  if qa[f].Equal(b.qa[ff]) then begin
  eq := true;
  break;
  end; {if}
  end; {next ff}
  
  
  if not eq then begin
  WriteLn(fp, '\section{',qa[f].subj,'}');
  WriteLn(fp, qa[f].q);
     for fff := Low(anss) to High(anss) do begin
      if fff = qa[f].tru then WriteLn(fp, '\subsection{+}')
      else WriteLn(fp, '\subsection{}'); WriteLn(fp, qa[f].a[fff]); end;
  WriteLn(fp, '\subsection{', qa[f].tr, '}');
  WriteLn(fp, qa[f].sol);
  end; {if}
  
  
  end; {next f}
Close(fp);
end;

function TBank.SaveToFile(fn : UTF8string) : boolean;
  var
      f, fff  : LongInt;
      fp : Text;
begin
Assign(fp, fn);
ReWrite(fp);
for f := Low(qa) to High(qa) do begin
  WriteLn(fp, '\section{',qa[f].subj,'}');
  WriteLn(fp, qa[f].q);
  for fff := Low(anss) to High(anss) do begin
      if fff = qa[f].tru then WriteLn(fp, '\subsection{+}')
      else WriteLn(fp, '\subsection{}'); WriteLn(fp, qa[f].a[fff]); end;
  WriteLn(fp, '\subsection{', qa[f].tr, '}');
  WriteLn(fp, qa[f].sol);
end; {next f}
Close(fp);
end;

function TBank.SaveToFile1Part(fn : UTF8string) : boolean;
  var
      f, fff  : LongInt;
      fp : Text;
      flots : array[subjs] of LongInt; // количество задач в пакете на соответствующую тему
begin
flots := lots;
for f := Low(flots) to High(flots) do flots[f] := flots[f] * (variants_count+1);
Assign(fp, fn);
ReWrite(fp);
for f := Low(qa) to High(qa) do begin
if flots[qa[f].subj] = 0 then continue else Dec(flots[qa[f].subj]);

  WriteLn(fp, '\section{',qa[f].subj,'}');
  WriteLn(fp, qa[f].q);
  for fff := Low(anss) to High(anss) do begin
      if fff = qa[f].tru then WriteLn(fp, '\subsection{+}')
      else WriteLn(fp, '\subsection{}'); WriteLn(fp, qa[f].a[fff]); end;
  WriteLn(fp, '\subsection{', qa[f].tr, '}');
  WriteLn(fp, qa[f].sol);
end; {next f}
Close(fp);
end;

procedure ErrCL; begin WriteLn(stderr, 'Ошибка параметров командной строки'); HALT; end;
procedure CLC(c : LongInt); begin if ParamCount = c then exit;
WriteLn(stderr, 'Число параметров не соответствует указанному ключу'); ErrCl; end;

  var
      f, ff, fff : LongInt;
      s  : utf8string;
      texlog : Text;
      b  : TBank;

begin

b  := TBank.Create;

if ParamStr(1) = '--stat' then begin
CLC(2);

end else begin
  if ParamStr(1) = '--compare' then begin
  CLC(3);
  with TBank.Create do begin
  LoadFromFile(ParamStr(2)); b.LoadFromFile(ParamStr(3));
  fff := 0;
  for f := Low(qa) to High(qa) do for ff := Low(b.qa) to High(b.qa) do begin
    if qa[f].Equal(b.qa[ff]) then begin
      WriteLn(f, #9, ff);
      inc(fff);
      break;
    end;
  end; {next f}
  WriteLn('Equals: ', fff);
  end; {w}
  end else begin
    if ParamStr(1) = '--exclude' then begin
    CLC(4);
      with TBank.Create do begin
      LoadFromFile(ParamStr(2)); b.LoadFromFile(ParamStr(3));
      ExcludeToFile(ParamStr(4), b);
      end;
    end else begin
      if ParamStr(1) = '--neat' then begin
      CLC(3);
         b.LoadFromFile(ParamStr(2));
         b.SaveToFile(ParamStr(3));
      end else begin
        if ParamStr(1) = '--getone' then begin
        CLC(3);
        b.LoadFromFile(ParamStr(2));
        b.SaveToFile1Part(ParamStr(3));
        end else begin
        
        
        end;
      end;
    end;
  end;
end;

WriteLn(stderr, 'УСПЕШНО завершено !');

end.